@@ -1,18 +1,24 @@
 import errno
+import io
 import logging
 import os
 import random
 import re
-import signal
 import socket
-import subprocess
+import stat
 import sys
 import threading
 import time
 import traceback
 import unittest
 
+try:
+    import configparser
+except ImportError:
+    import ConfigParser as configparser
+
 import psutil
+import subprocess32 as subprocess
 
 import mitogen.core
 import mitogen.fork
@@ -63,6 +69,67 @@ if faulthandler is not None:
 mitogen.core.LOG.propagate = True
 
 
+def base_executable(executable=None):
+    '''Return the path of the Python executable used to create the virtualenv.
+    '''
+    # https://docs.python.org/3/library/venv.html
+    # https://github.com/pypa/virtualenv/blob/main/src/virtualenv/discovery/py_info.py
+    # https://virtualenv.pypa.io/en/16.7.9/reference.html#compatibility-with-the-stdlib-venv-module
+    if executable is None:
+        executable = sys.executable
+
+    if not executable:
+        raise ValueError
+
+    try:
+        base_executable = sys._base_executable
+    except AttributeError:
+        base_executable = None
+
+    if base_executable and base_executable != executable:
+        return 'be', base_executable
+
+    # Python 2.x only has sys.base_prefix if running outside a virtualenv.
+    try:
+        sys.base_prefix
+    except AttributeError:
+        # Python 2.x outside a virtualenv
+        return executable
+
+    # Python 3.3+ has sys.base_prefix. In a virtualenv it differs to sys.prefix.
+    if sys.base_prefix == sys.prefix:
+        return executable
+
+    while executable.startswith(sys.prefix) and stat.S_ISLNK(os.lstat(executable).st_mode):
+        dirname = os.path.dirname(executable)
+        target = os.path.join(dirname, os.readlink(executable))
+        executable = os.path.abspath(os.path.normpath(target))
+        print(executable)
+
+    if executable.startswith(sys.base_prefix):
+        return executable
+
+    # Virtualenvs record details in pyvenv.cfg
+    parser = configparser.RawConfigParser()
+    with io.open(os.path.join(sys.prefix, 'pyvenv.cfg'), encoding='utf-8') as f:
+        content = u'[virtualenv]\n' + f.read()
+    try:
+        parser.read_string(content)
+    except AttributeError:
+        parser.readfp(io.StringIO(content))
+
+    # virtualenv style pyvenv.cfg includes the base executable.
+    # venv style pyvenv.cfg doesn't.
+    try:
+        return parser.get(u'virtualenv', u'base-executable')
+    except configparser.NoOptionError:
+        pass
+
+    basename = os.path.basename(executable)
+    home = parser.get(u'virtualenv', u'home')
+    return os.path.join(home, basename)
+
+
 def data_path(suffix):
     path = os.path.join(DATA_DIR, suffix)
     if path.endswith('.key'):
@@ -71,30 +138,6 @@ def data_path(suffix):
     return path
 
 
-def subprocess__check_output(*popenargs, **kwargs):
-    # Missing from 2.6.
-    process = subprocess.Popen(stdout=subprocess.PIPE, *popenargs, **kwargs)
-    output, _ = process.communicate()
-    retcode = process.poll()
-    if retcode:
-        cmd = kwargs.get("args")
-        if cmd is None:
-            cmd = popenargs[0]
-        raise subprocess.CalledProcessError(retcode, cmd)
-    return output
-
-
-def Popen__terminate(proc):
-    os.kill(proc.pid, signal.SIGTERM)
-
-
-if hasattr(subprocess, 'check_output'):
-    subprocess__check_output = subprocess.check_output
-
-if hasattr(subprocess.Popen, 'terminate'):
-    Popen__terminate = subprocess.Popen.terminate
-
-
 def threading__thread_is_alive(thread):
     """Return whether the thread is alive (Python version compatibility shim).
 
@@ -457,7 +500,7 @@ def get_docker_host():
 
 class DockerizedSshDaemon(object):
     def _get_container_port(self):
-        s = subprocess__check_output(['docker', 'port', self.container_name])
+        s = subprocess.check_output(['docker', 'port', self.container_name])
         for line in s.decode().splitlines():
             m = self.PORT_RE.match(line)
             if not m:
@@ -472,7 +515,7 @@ class DockerizedSshDaemon(object):
 
     def start_container(self):
         try:
-            subprocess__check_output(['docker', '--version'])
+            subprocess.check_output(['docker', '--version'])
         except Exception:
             raise unittest.SkipTest('Docker binary is unavailable')
 
@@ -486,7 +529,7 @@ class DockerizedSshDaemon(object):
             '--name', self.container_name,
             self.image,
         ]
-        subprocess__check_output(args)
+        subprocess.check_output(args)
         self._get_container_port()
 
     def __init__(self, mitogen_test_distro=os.environ.get('MITOGEN_TEST_DISTRO', 'debian9')):
@@ -518,7 +561,7 @@ class DockerizedSshDaemon(object):
     def check_processes(self):
         args = ['docker', 'exec', self.container_name, 'ps', '-o', 'comm=']
         counts = {}
-        for comm in subprocess__check_output(args).decode().splitlines():
+        for comm in subprocess.check_output(args).decode().splitlines():
             comm = comm.strip()
             counts[comm] = counts.get(comm, 0) + 1
 
@@ -533,7 +576,7 @@ class DockerizedSshDaemon(object):
 
     def close(self):
         args = ['docker', 'rm', '-f', self.container_name]
-        subprocess__check_output(args)
+        subprocess.check_output(args)
 
 
 class BrokerMixin(object):
